import 'package:flutter/material.dart';
import 'dart:math' show max, min;
import 'IIIFImageSettings.dart';
import 'IIIFTile.dart';

class IIIFGrid extends StatefulWidget {
  final IIIFImageSettings settings;
  IIIFGrid({Key key, this.settings}) : super(key: key);

  @override
  _IIIFGridState createState() => _IIIFGridState(settings: settings);
}

class _IIIFGridState extends State<IIIFGrid> {
  final IIIFImageSettings settings;
  _IIIFGridState({this.settings});

  //
  // CORE PARAMS
  //
  final GlobalKey rootKey = GlobalKey();
  get maxRows => (settings.meta.height / settings.tileSize).ceilToDouble();
  get maxCols => (settings.meta.width / settings.tileSize).ceilToDouble();
  get tileSize => settings.tileSize;
  final double padding = 25;
  //

  //
  //
  //
  // RETREIVE BOX SIZE
  Size get size {
    RenderBox box = rootKey.currentContext.findRenderObject();
    return box.size;
  }

  //
  //
  //
  // RETREIVE BOX OFFSET
  Offset get offset {
    RenderBox box = rootKey.currentContext.findRenderObject();
    return box.localToGlobal(Offset.zero);
  }

  //
  //INTERACTION PARAMS
  //
  final Matrix4 _transform = Matrix4.identity();
  double cx;
  double cy;
  double interactionStartScale;
  double currentScale;
  double minScale;

  @override
  toString({DiagnosticLevel minLevel = DiagnosticLevel.debug}) {
    return '_IIIF_GridState()';
  }

  @override
  void initState() {
    //WidgetsBinding.instance.addPostFrameCallback(_checkScaleUpdates);
    super.initState();
    settings.notifier.addListener(() {
      _initialized = false;
    });
  }

  List<Widget> _getTiles(_) {
    // CURRENT SCALE
    double scale = _transform.getMaxScaleOnAxis();
    // VIEWPORT
    double vw = size.width;
    double vh = size.height;
    // VISIBLE PART OF THE IMAGE
    double vwImage = vw / scale;
    //double vhImage = vh / scale;
    //
    // CURRENT ZOOM LEVEL
    int zoom = max(1, (vwImage / vw).round());
    // STEPPED SCALING
    double zoomScale = scale * zoom;
    //
    // TILESIZE IN ORIGINAL IMAGE
    double tileSideImage = zoom * settings.tileSize;
    //
    // HOW MANY COLUMNS/ROWS DO WE NEED TO GET
    int cols = (vw / settings.tileSize / zoomScale).ceil() + 1;
    int rows = (vh / settings.tileSize / zoomScale).ceil() + 1;
    //
    // TRANSLATION IN RELATION TO ORIGINAL SIZED IMAGE
    int tx = (_transform.getTranslation().x / zoomScale).ceil();
    int ty = (_transform.getTranslation().y / zoomScale).ceil();
    //
    // LEFT TOP VISIBLE TILE
    int sx = max(0, (-tx / settings.tileSize).floor());
    int sy = max(0, (-ty / settings.tileSize).floor());
    //
    //
    List<Widget> list = [];
    //
    //
    for (int x = sx; x < sx + cols; x++) {
      for (int y = sy; y < sy + rows; y++) {
        IIIFTile tile = IIIFTile.validate(
          x: x.toDouble(),
          y: y.toDouble(),
          rawTileSize: tileSideImage,
          settings: settings,
          translation: _transform.getTranslation(),
          scale: zoomScale,
        );
        if (tile != null) list.add(tile);
      }
    }

    //print(list.length);
    return list;
  }

  void center() {
    // CENTERING
    if (settings.meta.width * currentScale < size.width) {
      _transform.translate((size.width - (settings.meta.width * currentScale)) * 0.5 / currentScale);
    }
    if (settings.meta.height * currentScale < size.height) {
      _transform.translate(0.0, (size.height - (settings.meta.height * currentScale)) * 0.5 / currentScale);
    }
  }

  void _onScaleUpdate({ScaleUpdateDetails scaleUpdateDetails}) {
    // GET CURRENT SCALE
    currentScale = _transform.getMaxScaleOnAxis();
    if (interactionStartScale == null) interactionStartScale = currentScale;
    //
    //
    double fx = scaleUpdateDetails != null ? scaleUpdateDetails.focalPoint.dx : 0.0;
    double fy = scaleUpdateDetails != null ? scaleUpdateDetails.focalPoint.dy : 0.0;
    //
    //
    if ((scaleUpdateDetails != null ? scaleUpdateDetails.scale : 1) == 1) {
      // SIMPLE DRAGGING
      double dx = fx - (cx ?? 0);
      double dy = fy - (cy ?? 0);
      _transform.translate(dx / currentScale, dy / currentScale);
    } else {
      // HANDLE SCALING
      currentScale = _transform.getMaxScaleOnAxis();
      if (interactionStartScale == null) interactionStartScale = currentScale;
      // RESET SCALE
      _transform.scale(1 / currentScale);
      // SET SCALE
      double nextScale = min(1, max(minScale, scaleUpdateDetails.scale * interactionStartScale));
      _transform.scale(nextScale);
      // CALCULATE THE OFFSET TO ENABLE CENTERED ZOOMING
      double gx = (fx - _transform.getTranslation().x - offset.dx);
      double gy = (fy - _transform.getTranslation().y - offset.dy);
      _transform.translate(gx / nextScale - gx / currentScale, gy / nextScale - gy / currentScale);
    }

    // UPDATE POINTER POSITION ON SCREEN
    cx = fx;
    cy = fy;

    // LIMITS RIGHT / BOTTOM
    double rx = max(size.width - (_transform.getTranslation().x + settings.meta.width * currentScale), 0.0);
    double ry = max(size.height - (_transform.getTranslation().y + settings.meta.height * currentScale), 0.0);
    // if (rx < size.width)
    _transform.translate(rx / currentScale, ry / currentScale);

    // LIMITS LEFT / TOP
    double lx = min(0.0, _transform.getTranslation().x * -1);
    double ly = min(0.0, _transform.getTranslation().y * -1);
    _transform.translate(lx / currentScale, ly / currentScale, 0);

    center();

    // DONE
    setState(() {});
    //print(size);
  }

  bool _initialized = false;
  void initialize() {
    if (_initialized == false) {
      minScale = min((size.width - padding * 2) / settings.meta.width, (size.height - padding * 2) / settings.meta.height);
      _transform.scale(minScale);
      _onScaleUpdate();
      _initialized = true;
    }
  }

  @override
  Widget build(BuildContext context) {
    if (rootKey.currentContext == null) {
      Future.delayed(Duration(microseconds: 10), () {
        setState(() {});
        //
      });
      return Container(
        key: rootKey,
        color: Colors.blue,
      );
    }

    initialize();

    //
    //
    //
    Stack stack = Stack(
      children: _getTiles(null),
    );
    //
    //
    //
    return GestureDetector(
        onScaleStart: (scaleStartDetails) {
          // UPDATE POINTER POSITION ON SCREEN
          cx = scaleStartDetails.focalPoint.dx;
          cy = scaleStartDetails.focalPoint.dy;
        },
        onScaleUpdate: (scaleUpdateDetails) {
          _onScaleUpdate(scaleUpdateDetails: scaleUpdateDetails);
        },
        onScaleEnd: (scaleEndDetails) {
          interactionStartScale = null;
        },
        child: Stack(children: <Widget>[
          Container(
            key: rootKey,
            color: Colors.grey.withAlpha(0),
          ),
          stack,
          Container(
            transform: _transform,
            color: Color.fromARGB(100, 255, 0, 0),
            width: 0,
            height: 0,
          ),
        ]));
  }
}
